#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import logging
import os
import subprocess
import sys

from dbnd_airflow.utils import create_dbnd_pool


# DO NOT IMPORT ANYTHING FROM AIRFLOW
# we need to initialize some config values first
from dbnd import dbnd_config  # isort:skip
from dbnd._core.context.bootstrap import dbnd_system_bootstrap  # isort:skip


def subprocess_airflow(args):
    """Forward arguments to airflow command line"""

    from airflow.configuration import conf
    from sqlalchemy.engine.url import make_url

    # let's make sure that we user correct connection string
    airflow_sql_conn = conf.get("core", "SQL_ALCHEMY_CONN")
    env = os.environ.copy()
    env["AIRFLOW__CORE__SQL_ALCHEMY_CONN"] = airflow_sql_conn
    env["AIRFLOW__CORE__FERNET_KEY"] = conf.get("core", "FERNET_KEY")

    # if we use airflow, we can get airflow from external env
    args = [sys.executable, "-m", "dbnd_airflow"] + args
    logging.info(
        "Running airflow command at subprocess: '%s" " with DB=%s",
        subprocess.list2cmdline(args),
        repr(make_url(airflow_sql_conn)),
    )
    try:
        subprocess.check_call(args=args, env=env)
    except Exception:
        logging.error(
            "Failed to run airflow command %s with path=%s",
            subprocess.list2cmdline(args),
            sys.path,
        )
        raise
    logging.info("Airflow command has been successfully executed")


def subprocess_airflow_initdb():
    logging.info("Initializing Airflow DB")
    return subprocess_airflow(args=["initdb"])


def main(args=None):
    # from dbnd._core.log.config import configure_basic_logging
    # configure_basic_logging(None)

    dbnd_system_bootstrap()

    # LET'S PATCH AIRFLOW FIRST
    from dbnd_airflow.bootstrap import dbnd_airflow_bootstrap

    dbnd_airflow_bootstrap()

    from airflow.bin.cli import CLIFactory
    from airflow.configuration import conf
    from dbnd_airflow.plugins.setup_plugins import (
        setup_scheduled_dags,
        setup_versioned_dags,
    )

    # ORIGINAL CODE from  airflow/bin/airflow
    if conf.get("core", "security") == "kerberos":
        os.environ["KRB5CCNAME"] = conf.get("kerberos", "ccache")
        os.environ["KRB5_KTNAME"] = conf.get("kerberos", "keytab")

    import argcomplete
    from dbnd_airflow.scheduler.single_dag_run_job import find_and_kill_zombies

    CLIFactory.subparsers_dict[find_and_kill_zombies.__name__] = {
        "func": find_and_kill_zombies,
        "help": "Clean up BackfillJob zombie tasks",
        "args": tuple(),
    }

    parser = CLIFactory.get_parser()
    argcomplete.autocomplete(parser)
    args = parser.parse_args(args=args)
    func_name = args.func.__name__

    # DBND PATCH:
    if dbnd_config.getboolean("airflow", "auto_add_scheduled_dags") and func_name in [
        "scheduler",
        "webserver",
    ]:
        setup_scheduled_dags()
    if dbnd_config.getboolean("airflow", "auto_add_versioned_dags") and func_name in [
        "webserver"
    ]:
        setup_versioned_dags()

    args.func(args)

    if func_name in ["resetdb", "initdb"]:
        pool_name = dbnd_config.get("airflow", "dbnd_pool")
        if pool_name == "dbnd_pool":
            create_dbnd_pool(pool_name)


if __name__ == "__main__":
    main()



from datetime import timedelta

from airflow.models import Variable


def get_template_context(execution_date):
    from airflow import macros

    ds = execution_date.strftime("%Y-%m-%d")
    ts = execution_date.isoformat()
    yesterday_ds = (execution_date - timedelta(1)).strftime("%Y-%m-%d")
    tomorrow_ds = (execution_date + timedelta(1)).strftime("%Y-%m-%d")

    ds_nodash = ds.replace("-", "")
    ts_nodash = execution_date.strftime("%Y%m%dT%H%M%S")
    ts_nodash_with_tz = ts.replace("-", "").replace(":", "")
    yesterday_ds_nodash = yesterday_ds.replace("-", "")
    tomorrow_ds_nodash = tomorrow_ds.replace("-", "")

    class VariableAccessor:
        """
        Wrapper around Variable. This way you can get variables in templates by using
        {var.value.your_variable_name}.
        """

        def __init__(self):
            self.var = None

        def __getattr__(self, item):
            self.var = Variable.get(item)
            return self.var

        def __repr__(self):
            return str(self.var)

    class VariableJsonAccessor:
        """
        Wrapper around deserialized Variables. This way you can get variables
        in templates by using {var.json.your_variable_name}.
        """

        def __init__(self):
            self.var = None

        def __getattr__(self, item):
            self.var = Variable.get(item, deserialize_json=True)
            return self.var

        def __repr__(self):
            return str(self.var)

    return {
        "ds": ds,
        "ds_nodash": ds_nodash,
        "ts": ts,
        "ts_nodash": ts_nodash,
        "ts_nodash_with_tz": ts_nodash_with_tz,
        "yesterday_ds": yesterday_ds,
        "yesterday_ds_nodash": yesterday_ds_nodash,
        "tomorrow_ds": tomorrow_ds,
        "tomorrow_ds_nodash": tomorrow_ds_nodash,
        "END_DATE": ds,
        "end_date": ds,
        "execution_date": execution_date,
        "latest_date": ds,
        "macros": macros,
        "var": {"value": VariableAccessor(), "json": VariableJsonAccessor()},
    }



import logging


logger = logging.getLogger("dbnd-scheduler")


try:
    from dbnd_airflow.scheduler.scheduler_dags_provider import get_dags

    # airflow will only scan files containing the text DAG or airflow. This comment performs this function
    dags = get_dags()
    if dags:
        for dag in dags:
            globals()[dag.dag_id] = dag
except Exception as e:
    logging.exception("Failed to get dags form databand server")
    raise e



import logging

from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.tracking.no_tracking import should_not_track
from dbnd._core.utils.type_check_utils import is_instance_by_class_name
from dbnd_airflow.tracking.execute_tracking import new_execute


logger = logging.getLogger(__name__)


def _track_task(task):
    if should_not_track(task):
        return

    if is_instance_by_class_name(task, "SubDagOperator"):
        # we do not track the execute of a SubDag, only its tasks
        track_dag(task.subdag)
    else:
        track_operator(task)


def new_execute_for_class(self, context):
    return new_execute(context)


def track_operator(operator):
    import inspect

    if inspect.isclass(operator):
        from airflow.models import BaseOperator
        from airflow.operators.subdag_operator import SubDagOperator

        # we only track operators which base on Airflow BaseOperator
        if not issubclass(operator, BaseOperator):
            return operator

        # we are not tracking sub dags through this mechanism
        if issubclass(operator, SubDagOperator):
            return operator

        # the operator class is already tracked
        if (
            hasattr(operator, "_tracked_class")
            and operator.__name__ == operator._tracked_class
        ):
            return operator

        # this is the first time we encounter this class so we mark it
        operator._tracked_class = operator.__name__
        operator.__execute__ = operator.execute
        operator.execute = new_execute_for_class
        return operator

    else:
        # the operator instance is already tracked
        if hasattr(operator, "_tracked_instance"):
            return operator

        # this is the first time we encounter this instance so we mark it
        operator._tracked_instance = True

        # if the operator's class is tracked, we can't used `operator.__class__.execute`
        # need to use the original execute (__execute__)
        if (
            hasattr(operator, "_tracked_class")
            and operator.__class__.__name__ == operator._tracked_class
        ):
            operator.__execute__ = operator.__class__.__execute__
        else:
            # this can be a problem when the operator class doesn't implement it's own execute
            operator.__execute__ = operator.__class__.execute

        operator.execute = new_execute

    return operator


def _is_verbose():
    config = get_dbnd_project_config()
    return config.is_verbose()


def track_task(task):
    try:
        _track_task(task)
    except Exception:
        if _is_verbose():
            logger.exception("Failed to modify %s for tracking" % task.task_id)
        else:
            logger.info("Failed to modify %s for tracking" % task.task_id)


def track_dag(dag):
    """
    Modify operators in dag if necessary so we could track them.
    Supported operators: EmrAddStepsOperator, DataProcPySparkOperator, SparkSubmitOperator, PythonOperator.
    Other operators are not modified.

    More details on each operator:
    - EmrAddStepsOperator: modify args of spark-submit steps by adding dbnd variables using --conf
    - DataProcPySparkOperator: add dbnd variables to dataproc properties
    - SparkSubmitOperator: add dbnd variables to spark-submit command using --conf
    - PythonOperator: wrap python function with @task
    """
    try:
        if should_not_track(dag):
            return

        for task in dag.tasks:
            track_task(task)
    except Exception:
        logger.exception("Failed to modify %s for tracking" % dag.dag_id)



from airflow.utils.state import State

from dbnd._core.constants import TaskRunState
from dbnd._core.utils.timezone import utcnow
from dbnd_airflow.airflow_extensions.dal import update_airflow_task_instance_in_db


class AirflowTaskInstanceRetryController(object):
    def __init__(self, task_instance, task_run):
        self.task_instance = task_instance
        self.task_run = task_run
        pass

    def schedule_task_instance_for_retry(
        self, retry_count, retry_delay, increment_try_number
    ):

        if self._is_eligible_to_retry(self.task_instance):
            self.task_run.task.task_retries = retry_count
            self.task_run.task.task_retry_delay = retry_delay
            self.task_instance.state = State.UP_FOR_RETRY
            # Ensure that end date has a value. If it does not - airflow crashes when calculating next retry datetime
            self.task_instance.end_date = self.task_instance.end_date or utcnow()
            if increment_try_number:
                self.task_instance._try_number += 1
            update_airflow_task_instance_in_db(self.task_instance)
            self.task_run.set_task_run_state(TaskRunState.UP_FOR_RETRY, track=True)
            return True
        else:
            return False

    @staticmethod
    def _is_eligible_to_retry(task_instance):
        """Is task instance is eligible for retry"""
        return task_instance.try_number <= task_instance.max_tries



import logging
import re
import subprocess


logger = logging.getLogger(__name__)


def print_stack_trace(stack_frame):
    try:
        import traceback

        traceback.print_stack(stack_frame)
    except Exception as e:
        logger.info("Could not print stack trace! Exception: %s", e)


def print_cpu_memory_usage():
    try:
        import psutil

        cpu_usage_percent = psutil.cpu_percent(interval=1)
        virtual_memory = psutil.virtual_memory()
        last_minute_load, last_5_minute_load, last_15_minute_load = [
            x / psutil.cpu_count() * 100 for x in psutil.getloadavg()
        ]

        logger.info(
            """
                    Cpu usage %%: %s"
                    "Virtual memory: %s"
                    "Last minute cpu load %%: %s"
                    "Last 5 minute cpu load %%: %s"
                    "Last 15 minute cpu load %%: %s"
                    """
            % (
                cpu_usage_percent,
                virtual_memory,
                last_minute_load,
                last_5_minute_load,
                last_15_minute_load,
            )
        )
    except Exception as e:
        logger.info("Could not read cpu and memory usage! Exception: %s", e)


def print_dmesg():
    try:
        human_dmesg()
    except Exception as e:
        logger.info("Could not get dmesg data! Exception: %s", e)


_datetime_format = "%Y-%m-%d %H:%M:%S"
_dmesg_line_regex = re.compile("^\[(?P<time>\d+\.\d+)\](?P<line>.*)$")


def human_dmesg():
    from datetime import datetime, timedelta

    now = datetime.now()
    uptime_diff = None
    with open("/proc/uptime") as f:
        uptime_diff = f.read().strip().split()[0]
    uptime = now - timedelta(
        seconds=int(uptime_diff.split(".")[0]),
        microseconds=int(uptime_diff.split(".")[1]),
    )

    dmesg_data = subprocess.check_output(["dmesg"]).decode()
    for line in dmesg_data.split("\n"):
        if not line:
            continue
        match = _dmesg_line_regex.match(line)
        if match:
            seconds = int(match.groupdict().get("time", "").split(".")[0])
            nanoseconds = int(match.groupdict().get("time", "").split(".")[1])
            microseconds = int(round(nanoseconds * 0.001))
            line = match.groupdict().get("line", "")
            t = uptime + timedelta(seconds=seconds, microseconds=microseconds)
            logger.info("[%s]%s" % (t.strftime(_datetime_format), line))
